@@ -19,48 +19,21 @@
  * Boston, MA 02110-1301, USA.
  */
 
-#ifndef __GSSDP_CLIENT_H__
-#define __GSSDP_CLIENT_H__
+#ifndef GSSDP_CLIENT_H
+#define GSSDP_CLIENT_H
 
 #include <glib-object.h>
 #include <gio/gio.h>
 
 G_BEGIN_DECLS
 
-GType
-gssdp_client_get_type (void) G_GNUC_CONST;
+#define GSSDP_TYPE_CLIENT (gssdp_client_get_type ())
 
-#define GSSDP_TYPE_CLIENT \
-                (gssdp_client_get_type ())
-#define GSSDP_CLIENT(obj) \
-                (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
-                 GSSDP_TYPE_CLIENT, \
-                 GSSDPClient))
-#define GSSDP_CLIENT_CLASS(klass) \
-                (G_TYPE_CHECK_CLASS_CAST ((klass), \
-                 GSSDP_TYPE_CLIENT, \
-                 GSSDPClientClass))
-#define GSSDP_IS_CLIENT(obj) \
-                (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
-                 GSSDP_TYPE_CLIENT))
-#define GSSDP_IS_CLIENT_CLASS(klass) \
-                (G_TYPE_CHECK_CLASS_TYPE ((klass), \
-                 GSSDP_TYPE_CLIENT))
-#define GSSDP_CLIENT_GET_CLASS(obj) \
-                (G_TYPE_INSTANCE_GET_CLASS ((obj), \
-                 GSSDP_TYPE_CLIENT, \
-                 GSSDPClientClass))
+G_DECLARE_DERIVABLE_TYPE (GSSDPClient, gssdp_client, GSSDP, CLIENT, GObject)
 
-typedef struct _GSSDPClientPrivate GSSDPClientPrivate;
 typedef struct _GSSDPClient GSSDPClient;
 typedef struct _GSSDPClientClass GSSDPClientClass;
 
-struct _GSSDPClient {
-        GObject parent;
-
-        GSSDPClientPrivate *priv;
-};
-
 struct _GSSDPClientClass {
         GObjectClass parent_class;
 
@@ -72,8 +45,7 @@ struct _GSSDPClientClass {
 };
 
 GSSDPClient *
-gssdp_client_new              (GMainContext *main_context,
-                               const char   *iface,
+gssdp_client_new              (const char   *iface,
                                GError      **error);
 
 GSSDPClient *
@@ -81,11 +53,6 @@ gssdp_client_new_with_port    (const char *iface,
                                guint16     msearch_port,
                                GError    **error);
 
-#ifndef GSSDP_DISABLE_DEPRECATED
-GMainContext *
-gssdp_client_get_main_context (GSSDPClient  *client);
-#endif
-
 void
 gssdp_client_set_server_id    (GSSDPClient  *client,
                                const char   *server_id);
@@ -109,8 +76,14 @@ gssdp_client_get_network      (GSSDPClient  *client);
 gboolean
 gssdp_client_get_active       (GSSDPClient  *client);
 
-GInetAddressMask *
-gssdp_client_get_address_mask (GSSDPClient *client);
+GInetAddress *
+gssdp_client_get_address      (GSSDPClient *client);
+
+guint
+gssdp_client_get_index        (GSSDPClient *client);
+
+GSocketFamily
+gssdp_client_get_family       (GSSDPClient *client);
 
 void
 gssdp_client_append_header    (GSSDPClient *client,
@@ -135,4 +108,4 @@ gssdp_client_guess_user_agent (GSSDPClient *client,
 
 G_END_DECLS
 
-#endif /* __GSSDP_CLIENT_H__ */
+#endif /* GSSDP_CLIENT_H */
