@@ -30,25 +30,23 @@
  * automatically when activating the browser.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif /* HAVE_CONFIG_H */
+
+#include "gssdp-resource-browser.h"
+#include "gssdp-client-private.h"
+#include "gssdp-protocol.h"
+
 #include <libsoup/soup.h>
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 
-#include "gssdp-resource-browser.h"
-#include "gssdp-client-private.h"
-#include "gssdp-protocol.h"
-#include "gssdp-marshal.h"
-
 #define RESCAN_TIMEOUT 5 /* 5 seconds */
 #define MAX_DISCOVERY_MESSAGES 3
 #define DISCOVERY_FREQUENCY    500 /* 500 ms */
 
-G_DEFINE_TYPE (GSSDPResourceBrowser,
-               gssdp_resource_browser,
-               G_TYPE_OBJECT);
-
 struct _GSSDPResourceBrowserPrivate {
         GSSDPClient *client;
 
@@ -70,6 +68,11 @@ struct _GSSDPResourceBrowserPrivate {
         GSource     *refresh_cache_src;
         GHashTable  *fresh_resources;
 };
+typedef struct _GSSDPResourceBrowserPrivate GSSDPResourceBrowserPrivate;
+
+G_DEFINE_TYPE_WITH_PRIVATE (GSSDPResourceBrowser,
+                            gssdp_resource_browser,
+                            G_TYPE_OBJECT);
 
 enum {
         PROP_0,
@@ -126,14 +129,12 @@ resource_unavailable             (GSSDPResourceBrowser *resource_browser,
 static void
 gssdp_resource_browser_init (GSSDPResourceBrowser *resource_browser)
 {
-        resource_browser->priv = G_TYPE_INSTANCE_GET_PRIVATE
-                                        (resource_browser,
-                                         GSSDP_TYPE_RESOURCE_BROWSER,
-                                         GSSDPResourceBrowserPrivate);
+        GSSDPResourceBrowserPrivate *priv = NULL;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
-        resource_browser->priv->mx = SSDP_DEFAULT_MX;
+        priv->mx = SSDP_DEFAULT_MX;
 
-        resource_browser->priv->resources =
+        priv->resources =
                 g_hash_table_new_full (g_str_hash,
                                        g_str_equal,
                                        g_free,
@@ -214,22 +215,24 @@ static void
 gssdp_resource_browser_dispose (GObject *object)
 {
         GSSDPResourceBrowser *resource_browser;
+        GSSDPResourceBrowserPrivate *priv;
 
         resource_browser = GSSDP_RESOURCE_BROWSER (object);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
-        if (resource_browser->priv->client) {
+        if (priv->client) {
                 if (g_signal_handler_is_connected
-                        (resource_browser->priv->client,
-                         resource_browser->priv->message_received_id)) {
+                        (priv->client,
+                         priv->message_received_id)) {
                         g_signal_handler_disconnect
-                                (resource_browser->priv->client,
-                                 resource_browser->priv->message_received_id);
+                                (priv->client,
+                                 priv->message_received_id);
                 }
 
                 stop_discovery (resource_browser);
 
-                g_object_unref (resource_browser->priv->client);
-                resource_browser->priv->client = NULL;
+                g_object_unref (priv->client);
+                priv->client = NULL;
         }
 
         clear_cache (resource_browser);
@@ -241,15 +244,16 @@ static void
 gssdp_resource_browser_finalize (GObject *object)
 {
         GSSDPResourceBrowser *resource_browser;
+        GSSDPResourceBrowserPrivate *priv;
 
         resource_browser = GSSDP_RESOURCE_BROWSER (object);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
-        if (resource_browser->priv->target_regex)
-                g_regex_unref (resource_browser->priv->target_regex);
+        g_clear_pointer (&priv->target_regex, g_regex_unref);
 
-        g_free (resource_browser->priv->target);
+        g_free (priv->target);
 
-        g_hash_table_destroy (resource_browser->priv->resources);
+        g_hash_table_destroy (priv->resources);
 
         G_OBJECT_CLASS (gssdp_resource_browser_parent_class)->finalize (object);
 }
@@ -266,8 +270,6 @@ gssdp_resource_browser_class_init (GSSDPResourceBrowserClass *klass)
         object_class->dispose      = gssdp_resource_browser_dispose;
         object_class->finalize     = gssdp_resource_browser_finalize;
 
-        g_type_class_add_private (klass, sizeof (GSSDPResourceBrowserPrivate));
-
         /**
          * GSSDPResourceBrowser:client:
          *
@@ -357,8 +359,7 @@ gssdp_resource_browser_class_init (GSSDPResourceBrowserClass *klass)
                               G_SIGNAL_RUN_LAST,
                               G_STRUCT_OFFSET (GSSDPResourceBrowserClass,
                                                resource_available),
-                              NULL, NULL,
-                              gssdp_marshal_VOID__STRING_POINTER,
+                              NULL, NULL, NULL,
                               G_TYPE_NONE,
                               2,
                               G_TYPE_STRING,
@@ -379,8 +380,7 @@ gssdp_resource_browser_class_init (GSSDPResourceBrowserClass *klass)
                               G_SIGNAL_RUN_LAST,
                               G_STRUCT_OFFSET (GSSDPResourceBrowserClass,
                                                resource_unavailable),
-                              NULL, NULL,
-                              gssdp_marshal_VOID__STRING,
+                              NULL, NULL, NULL,
                               G_TYPE_NONE,
                               1,
                               G_TYPE_STRING);
@@ -424,13 +424,17 @@ static void
 gssdp_resource_browser_set_client (GSSDPResourceBrowser *resource_browser,
                                    GSSDPClient          *client)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser));
         g_return_if_fail (GSSDP_IS_CLIENT (client));
 
-        resource_browser->priv->client = g_object_ref (client);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
-        resource_browser->priv->message_received_id =
-                g_signal_connect_object (resource_browser->priv->client,
+        priv->client = g_object_ref (client);
+
+        priv->message_received_id =
+                g_signal_connect_object (priv->client,
                                          "message-received",
                                          G_CALLBACK (message_received_cb),
                                          resource_browser,
@@ -448,10 +452,14 @@ gssdp_resource_browser_set_client (GSSDPResourceBrowser *resource_browser,
 GSSDPClient *
 gssdp_resource_browser_get_client (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_val_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser),
                               NULL);
 
-        return resource_browser->priv->client;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        return priv->client;
 }
 
 /**
@@ -469,16 +477,18 @@ gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
         char *version;
         const char *version_pattern;
         GError *error;
+        GSSDPResourceBrowserPrivate *priv;
 
         g_return_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser));
         g_return_if_fail (target != NULL);
-        g_return_if_fail (!resource_browser->priv->active);
-        
-        g_free (resource_browser->priv->target);
-        resource_browser->priv->target = g_strdup (target);
 
-        if (resource_browser->priv->target_regex)
-                g_regex_unref (resource_browser->priv->target_regex);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        g_return_if_fail (!priv->active);
+
+        g_free (priv->target);
+        priv->target = g_strdup (target);
+
+        g_clear_pointer (&priv->target_regex, g_regex_unref);
 
         version_pattern = "([0-9]+)";
         /* Make sure we have enough room for version pattern */
@@ -493,15 +503,15 @@ gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
                                   version + 1,
                                   G_REGEX_ANCHORED,
                                   G_REGEX_MATCH_ANCHORED)) {
-                resource_browser->priv->version = atoi (version + 1);
+                priv->version = atoi (version + 1);
                 strcpy (version + 1, version_pattern);
         }
 
         error = NULL;
-        resource_browser->priv->target_regex = g_regex_new (pattern,
-                                                            0,
-                                                            0,
-                                                            &error);
+        priv->target_regex = g_regex_new (pattern,
+                                          0,
+                                          0,
+                                          &error);
         if (error) {
                 g_warning ("Error compiling regular expression '%s': %s",
                            pattern,
@@ -523,10 +533,12 @@ gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
 const char *
 gssdp_resource_browser_get_target (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
         g_return_val_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser),
                               NULL);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
-        return resource_browser->priv->target;
+        return priv->target;
 }
 
 /**
@@ -540,12 +552,16 @@ void
 gssdp_resource_browser_set_mx (GSSDPResourceBrowser *resource_browser,
                                gushort               mx)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser));
 
-        if (resource_browser->priv->mx == mx)
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        if (priv->mx == mx)
                 return;
 
-        resource_browser->priv->mx = mx;
+        priv->mx = mx;
         
         g_object_notify (G_OBJECT (resource_browser), "mx");
 }
@@ -559,9 +575,13 @@ gssdp_resource_browser_set_mx (GSSDPResourceBrowser *resource_browser,
 gushort
 gssdp_resource_browser_get_mx (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_val_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser), 0);
 
-        return resource_browser->priv->mx;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        return priv->mx;
 }
 
 /**
@@ -575,12 +595,15 @@ void
 gssdp_resource_browser_set_active (GSSDPResourceBrowser *resource_browser,
                                    gboolean              active)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser));
 
-        if (resource_browser->priv->active == active)
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        if (priv->active == active)
                 return;
 
-        resource_browser->priv->active = active;
+        priv->active = active;
 
         if (active) {
                 start_discovery (resource_browser);
@@ -602,9 +625,13 @@ gssdp_resource_browser_set_active (GSSDPResourceBrowser *resource_browser,
 gboolean
 gssdp_resource_browser_get_active (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_val_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser), 0);
 
-        return resource_browser->priv->active;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        return priv->active;
 }
 
 /**
@@ -619,11 +646,15 @@ gssdp_resource_browser_get_active (GSSDPResourceBrowser *resource_browser)
 gboolean
 gssdp_resource_browser_rescan (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
         g_return_val_if_fail (GSSDP_IS_RESOURCE_BROWSER (resource_browser), 0);
 
-        if (resource_browser->priv->active &&
-            resource_browser->priv->timeout_src == NULL &&
-            resource_browser->priv->refresh_cache_src == NULL) {
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        if (priv->active &&
+            priv->timeout_src == NULL &&
+            priv->refresh_cache_src == NULL) {
                 start_discovery (resource_browser);
                 return TRUE;
         }
@@ -638,12 +669,14 @@ static gboolean
 resource_expire (gpointer user_data)
 {
         GSSDPResourceBrowser *resource_browser;
+        GSSDPResourceBrowserPrivate *priv;
         Resource *resource;
         char *usn;
         char *canonical_usn;
 
         resource = user_data;
         resource_browser = resource->resource_browser;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
         /* Steal the USN pointer from the resource as we need it for the signal
          * emission.
@@ -651,7 +684,7 @@ resource_expire (gpointer user_data)
         usn = resource->usn;
         resource->usn = NULL;
 
-        if (resource_browser->priv->version > 0) {
+        if (priv->version > 0) {
                 char *version;
 
                 version = g_strrstr (usn, ":");
@@ -660,8 +693,7 @@ resource_expire (gpointer user_data)
                 canonical_usn = g_strdup (usn);
         }
 
-        g_hash_table_remove (resource->resource_browser->priv->resources,
-                             canonical_usn);
+        g_hash_table_remove (priv->resources, canonical_usn);
 
         g_signal_emit (resource_browser,
                        signals[RESOURCE_UNAVAILABLE],
@@ -673,10 +705,40 @@ resource_expire (gpointer user_data)
         return FALSE;
 }
 
+static char *
+rewrite_uri (const char *header, int index)
+{
+        const char *host = NULL;
+        SoupURI *uri = NULL;
+        GInetAddress *addr = NULL;
+        char *retval = NULL;
+
+        uri = soup_uri_new (header);
+        host = soup_uri_get_host (uri);
+        addr = g_inet_address_new_from_string (host);
+
+        if (g_inet_address_get_is_link_local (addr)) {
+                char *new_host;
+
+                new_host = g_strdup_printf ("%s%%%d",
+                                host,
+                                index);
+                soup_uri_set_host (uri, new_host);
+                g_free (new_host);
+        }
+
+        g_object_unref (addr);
+        retval = soup_uri_to_string (uri, FALSE);
+        soup_uri_free (uri);
+
+        return retval;
+}
+
 static void
 resource_available (GSSDPResourceBrowser *resource_browser,
                     SoupMessageHeaders   *headers)
 {
+        GSSDPResourceBrowserPrivate *priv;
         const char *usn;
         const char *header;
         Resource *resource;
@@ -686,25 +748,36 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         gboolean destroyLocations;
         GList *it1, *it2;
         char *canonical_usn;
+        GSocketFamily family = G_SOCKET_FAMILY_INVALID;
+        int index;
 
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
         usn = soup_message_headers_get_one (headers, "USN");
         if (!usn)
                 return; /* No USN specified */
 
+        family = gssdp_client_get_family (priv->client);
+        index = gssdp_client_get_index (priv->client);
+
         /* Build list of locations */
         locations = NULL;
         destroyLocations = TRUE;
 
         header = soup_message_headers_get_one (headers, "Location");
-        if (header)
-                locations = g_list_append (locations, g_strdup (header));
+        if (header) {
+                if (family == G_SOCKET_FAMILY_IPV6) {
+                        locations = g_list_append (locations, rewrite_uri (header, index));
+                } else {
+                        locations = g_list_append (locations, g_strdup (header));
+                }
+        }
 
         header = soup_message_headers_get_one (headers, "AL");
         if (header) {
                 /* Parse AL header. The format is:
                  * <uri1><uri2>... */
-                const char *start, *end;
-                char *uri;
+                const char *start;
+                char *end;
 
                 start = header;
                 while ((start = strchr (start, '<'))) {
@@ -716,8 +789,15 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                         if (!end || !*end)
                                 break;
 
-                        uri = g_strndup (start, end - start);
-                        locations = g_list_append (locations, uri);
+                        if (family == G_SOCKET_FAMILY_IPV6) {
+                                *end = '\0';
+                                locations = g_list_append (locations, rewrite_uri (start, index));
+                                *end = '>';
+                        } else {
+                                char *uri;
+                                uri = g_strndup (start, end - start);
+                                locations = g_list_append (locations, uri);
+                        }
 
                         start = end;
                 }
@@ -726,7 +806,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         if (!locations)
                 return; /* No location specified */
 
-        if (resource_browser->priv->version > 0) {
+        if (priv->version > 0) {
                 char *version;
 
                 version = g_strrstr (usn, ":");
@@ -736,12 +816,12 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         }
 
         /* Get from cache, if possible */
-        resource = g_hash_table_lookup (resource_browser->priv->resources,
+        resource = g_hash_table_lookup (priv->resources,
                                         canonical_usn);
         /* Put usn into fresh resources, so this resource will not be
          * removed on cache refreshing. */
-        if (resource_browser->priv->fresh_resources != NULL) {
-                g_hash_table_add (resource_browser->priv->fresh_resources,
+        if (priv->fresh_resources != NULL) {
+                g_hash_table_add (priv->fresh_resources,
                                   g_strdup (canonical_usn));
         }
 
@@ -775,7 +855,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                 resource->locations        = locations;
                 destroyLocations = FALSE; /* Ownership passed to resource */
                 
-                g_hash_table_insert (resource_browser->priv->resources,
+                g_hash_table_insert (priv->resources,
                                      canonical_usn,
                                      resource);
                 
@@ -785,8 +865,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                 canonical_usn = NULL;
         }
 
-        if (canonical_usn != NULL)
-                g_free (canonical_usn);
+        g_free (canonical_usn);
 
         /* Calculate new timeout */
         header = soup_message_headers_get_one (headers, "Cache-Control");
@@ -880,14 +959,16 @@ static void
 resource_unavailable (GSSDPResourceBrowser *resource_browser,
                       SoupMessageHeaders   *headers)
 {
+        GSSDPResourceBrowserPrivate *priv;
         const char *usn;
         char *canonical_usn;
 
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
         usn = soup_message_headers_get_one (headers, "USN");
         if (!usn)
                 return; /* No USN specified */
 
-        if (resource_browser->priv->version > 0) {
+        if (priv->version > 0) {
                 char *version;
 
                 version = g_strrstr (usn, ":");
@@ -897,11 +978,11 @@ resource_unavailable (GSSDPResourceBrowser *resource_browser,
         }
 
         /* Only process if we were cached */
-        if (!g_hash_table_lookup (resource_browser->priv->resources,
+        if (!g_hash_table_lookup (priv->resources,
                                   canonical_usn))
                 goto out;
 
-        g_hash_table_remove (resource_browser->priv->resources,
+        g_hash_table_remove (priv->resources,
                              canonical_usn);
 
         g_signal_emit (resource_browser,
@@ -917,15 +998,17 @@ static gboolean
 check_target_compat (GSSDPResourceBrowser *resource_browser,
                      const char           *st)
 {
+        GSSDPResourceBrowserPrivate *priv;
         GMatchInfo *info;
         int         version;
         char       *tmp;
 
-        if (strcmp (resource_browser->priv->target,
-                    GSSDP_ALL_RESOURCES) == 0)
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
+        if (g_str_equal (priv->target, GSSDP_ALL_RESOURCES))
                 return TRUE;
 
-        if (!g_regex_match (resource_browser->priv->target_regex,
+        if (!g_regex_match (priv->target_regex,
                             st,
                             0,
                             &info)) {
@@ -935,7 +1018,7 @@ check_target_compat (GSSDPResourceBrowser *resource_browser,
         }
 
         /* If there was no version to match, we're done */
-        if (resource_browser->priv->version == 0) {
+        if (priv->version == 0) {
                 g_match_info_free (info);
 
                 return TRUE;
@@ -955,7 +1038,7 @@ check_target_compat (GSSDPResourceBrowser *resource_browser,
             return FALSE;
         }
 
-        return (guint) version >= resource_browser->priv->version;
+        return (guint) version >= priv->version;
 }
 
 static void
@@ -1014,10 +1097,12 @@ message_received_cb (G_GNUC_UNUSED GSSDPClient *client,
                      gpointer                   user_data)
 {
         GSSDPResourceBrowser *resource_browser;
+        GSSDPResourceBrowserPrivate *priv;
 
         resource_browser = GSSDP_RESOURCE_BROWSER (user_data);
 
-        if (!resource_browser->priv->active)
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        if (!priv->active)
                 return;
 
         switch (type) {
@@ -1070,8 +1155,11 @@ clear_cache_helper (G_GNUC_UNUSED gpointer key,
 static void
 clear_cache (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
         /* Clear cache */
-        g_hash_table_foreach_remove (resource_browser->priv->resources,
+        g_hash_table_foreach_remove (priv->resources,
                                      clear_cache_helper,
                                      NULL);
 }
@@ -1080,48 +1168,66 @@ clear_cache (GSSDPResourceBrowser *resource_browser)
 static void
 send_discovery_request (GSSDPResourceBrowser *resource_browser)
 {
-        char *message;
+        GSSDPResourceBrowserPrivate *priv;
+        char *message = NULL;
+        const char *group = NULL;
+        char *dest = NULL;
 
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        group = _gssdp_client_get_mcast_group (priv->client);
+
+        /* FIXME: Check for IPv6 - ugly and remove strcpys */
+        if (strchr (group, ':')) {
+            dest = g_strdup_printf ("[%s]", group);
+        } else {
+            dest = g_strdup (group);
+        }
+
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
         message = g_strdup_printf (SSDP_DISCOVERY_REQUEST,
-                                   resource_browser->priv->target,
-                                   resource_browser->priv->mx,
+                                   dest,
+                                   priv->target,
+                                   priv->mx,
                                    g_get_prgname () ? g_get_prgname () : "");
 
-        _gssdp_client_send_message (resource_browser->priv->client,
+        _gssdp_client_send_message (priv->client,
                                     NULL,
                                     0,
                                     message,
                                     _GSSDP_DISCOVERY_REQUEST);
 
+        g_free (dest);
         g_free (message);
 }
 
 static gboolean
 discovery_timeout (gpointer data)
 {
+        GSSDPResourceBrowserPrivate *priv;
         GSSDPResourceBrowser *resource_browser;
 
         resource_browser = GSSDP_RESOURCE_BROWSER (data);
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
 
         send_discovery_request (resource_browser);
 
-        resource_browser->priv->num_discovery += 1;
+        priv->num_discovery += 1;
 
-        if (resource_browser->priv->num_discovery >= MAX_DISCOVERY_MESSAGES) {
-                resource_browser->priv->timeout_src = NULL;
-                resource_browser->priv->num_discovery = 0;
+        if (priv->num_discovery >= MAX_DISCOVERY_MESSAGES) {
+                priv->timeout_src = NULL;
+                priv->num_discovery = 0;
 
                 /* Setup cache refreshing */
-                resource_browser->priv->refresh_cache_src =
+                priv->refresh_cache_src =
                                   g_timeout_source_new_seconds (RESCAN_TIMEOUT);
                 g_source_set_callback
-                                     (resource_browser->priv->refresh_cache_src,
+                                     (priv->refresh_cache_src,
                                       refresh_cache,
                                       resource_browser,
                                       NULL);
-                g_source_attach (resource_browser->priv->refresh_cache_src,
+                g_source_attach (priv->refresh_cache_src,
                                  g_main_context_get_thread_default ());
-                g_source_unref (resource_browser->priv->refresh_cache_src);
+                g_source_unref (priv->refresh_cache_src);
 
                 return FALSE;
         } else
@@ -1132,24 +1238,28 @@ discovery_timeout (gpointer data)
 static void
 start_discovery (GSSDPResourceBrowser *resource_browser)
 {
+        GSSDPResourceBrowserPrivate *priv;
+
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+
         /* Send one now */
         send_discovery_request (resource_browser);
 
         /* And schedule the rest for later */
-        resource_browser->priv->num_discovery = 1;
-        resource_browser->priv->timeout_src =
+        priv->num_discovery = 1;
+        priv->timeout_src =
                 g_timeout_source_new (DISCOVERY_FREQUENCY);
-        g_source_set_callback (resource_browser->priv->timeout_src,
+        g_source_set_callback (priv->timeout_src,
                                discovery_timeout,
                                resource_browser, NULL);
 
-        g_source_attach (resource_browser->priv->timeout_src,
+        g_source_attach (priv->timeout_src,
                          g_main_context_get_thread_default ());
 
-        g_source_unref (resource_browser->priv->timeout_src);
+        g_source_unref (priv->timeout_src);
 
         /* Setup a set of responsive resources for cache refreshing */
-        resource_browser->priv->fresh_resources = g_hash_table_new_full
+        priv->fresh_resources = g_hash_table_new_full
                                         (g_str_hash,
                                          g_str_equal,
                                          g_free,
@@ -1160,19 +1270,16 @@ start_discovery (GSSDPResourceBrowser *resource_browser)
 static void
 stop_discovery (GSSDPResourceBrowser *resource_browser)
 {
-        if (resource_browser->priv->timeout_src) {
-                g_source_destroy (resource_browser->priv->timeout_src);
-                resource_browser->priv->timeout_src = NULL;
-                resource_browser->priv->num_discovery = 0;
-        }
-        if (resource_browser->priv->refresh_cache_src) {
-                g_source_destroy (resource_browser->priv->refresh_cache_src);
-                resource_browser->priv->refresh_cache_src = NULL;
-        }
-        if (resource_browser->priv->fresh_resources) {
-                g_hash_table_unref (resource_browser->priv->fresh_resources);
-                resource_browser->priv->fresh_resources = NULL;
+        GSSDPResourceBrowserPrivate *priv;
+
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        if (priv->timeout_src) {
+                priv->num_discovery = 0;
         }
+
+        g_clear_pointer (&priv->timeout_src, g_source_destroy);
+        g_clear_pointer (&priv->refresh_cache_src, g_source_destroy);
+        g_clear_pointer (&priv->fresh_resources, g_hash_table_destroy);
 }
 
 static gboolean
@@ -1201,14 +1308,16 @@ static gboolean
 refresh_cache (gpointer data)
 {
         GSSDPResourceBrowser *resource_browser;
+        GSSDPResourceBrowserPrivate *priv;
 
         resource_browser = GSSDP_RESOURCE_BROWSER (data);
-        g_hash_table_foreach_remove (resource_browser->priv->resources,
+        priv = gssdp_resource_browser_get_instance_private (resource_browser);
+        g_hash_table_foreach_remove (priv->resources,
                                      refresh_cache_helper,
-                                     resource_browser->priv->fresh_resources);
-        g_hash_table_unref (resource_browser->priv->fresh_resources);
-        resource_browser->priv->fresh_resources = NULL;
-        resource_browser->priv->refresh_cache_src = NULL;
+                                     priv->fresh_resources);
+        g_hash_table_unref (priv->fresh_resources);
+        priv->fresh_resources = NULL;
+        priv->refresh_cache_src = NULL;
 
         return FALSE;
 }
