@@ -19,7 +19,14 @@
  * Boston, MA 02110-1301, USA.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif /* HAVE_CONFIG_H */
+
+#include "gssdp-error.h"
+#include "gssdp-socket-functions.h"
+#include "gssdp-pktinfo-message.h"
+
 #include <errno.h>
 #include <string.h>
 
@@ -36,10 +43,7 @@
     #include <arpa/inet.h>
 #endif
 
-#include "gssdp-error.h"
-#include "gssdp-socket-functions.h"
-#include "gssdp-pktinfo-message.h"
-
+#include "gssdp-pktinfo6-message.h"
 static char*
 gssdp_socket_error_message (int error) {
 #ifdef G_OS_WIN32
@@ -93,20 +97,30 @@ gssdp_socket_option_set (GSocket    *socket,
 gboolean
 gssdp_socket_mcast_interface_set (GSocket      *socket,
                                   GInetAddress *iface_address,
+                                  gint          index,
                                   GError      **error) {
 
         const guint8 *address;
         gsize native_size;
 
-        address = g_inet_address_to_bytes (iface_address);
-        native_size = g_inet_address_get_native_size (iface_address);
+        if (g_inet_address_get_family (iface_address) == G_SOCKET_FAMILY_IPV6) {
+                return gssdp_socket_option_set (socket,
+                                                IPPROTO_IPV6,
+                                                IPV6_MULTICAST_IF,
+                                                (char *)&index,
+                                                sizeof (index),
+                                                error);
+        } else {
+                address = g_inet_address_to_bytes (iface_address);
+                native_size = g_inet_address_get_native_size (iface_address);
 
-        return gssdp_socket_option_set (socket,
-                                        IPPROTO_IP,
-                                        IP_MULTICAST_IF,
-                                        (char *) address,
-                                        native_size,
-                                        error);
+                return gssdp_socket_option_set (socket,
+                                                IPPROTO_IP,
+                                                IP_MULTICAST_IF,
+                                                (char *) address,
+                                                native_size,
+                                                error);
+        }
 }
 
 #define __GSSDP_UNUSED(x) (void)(x)
@@ -144,6 +158,14 @@ gssdp_socket_enable_info         (GSocket *socket,
         /* Register the type so g_socket_control_message_deserialize() will
          * find it */
         g_object_unref (g_object_new (GSSDP_TYPE_PKTINFO_MESSAGE, NULL));
+        g_object_unref (g_object_new (GSSDP_TYPE_PKTINFO6_MESSAGE, NULL));
+
+        gssdp_socket_option_set (socket,
+                                 IPPROTO_IPV6,
+                                 IPV6_RECVPKTINFO,
+                                 (char *)&enable,
+                                 sizeof (enable),
+                                 error);
 
         return gssdp_socket_option_set (socket,
                                         IPPROTO_IP,
