@@ -24,14 +24,39 @@
  * Boston, MA 02110-1301, USA.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
-#include <glib.h>
+#endif /* HAVE_CONFIG_H */
+#include <gio/gio.h>
 
 #include "gssdp-socket-functions.h"
 #include "gssdp-socket-source.h"
 #include "gssdp-protocol.h"
 #include "gssdp-error.h"
 
+#include <glib.h>
+
+struct _GSSDPSocketSource {
+        GObject parent;
+};
+
+struct _GSSDPSocketSourceClass {
+        GObjectClass parent_class;
+};
+
+struct _GSSDPSocketSourcePrivate {
+        GSource              *source;
+        GSocket              *socket;
+        GSSDPSocketSourceType type;
+
+        GInetAddress         *address;
+        char                 *device_name;
+        gint                  index;
+        guint                 ttl;
+        guint                 port;
+};
+typedef struct _GSSDPSocketSourcePrivate GSSDPSocketSourcePrivate;
+
 static void
 gssdp_socket_source_initable_init (gpointer g_iface,
                                    gpointer iface_data);
@@ -40,35 +65,23 @@ G_DEFINE_TYPE_EXTENDED (GSSDPSocketSource,
                         gssdp_socket_source,
                         G_TYPE_OBJECT,
                         0,
+                        G_ADD_PRIVATE (GSSDPSocketSource)
                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
                                     gssdp_socket_source_initable_init));
 
-struct _GSSDPSocketSourcePrivate {
-        GSource              *source;
-        GSocket              *socket;
-        GSSDPSocketSourceType type;
-
-        char                 *host_ip;
-        char                 *device_name;
-        guint                 ttl;
-        guint                 port;
-};
-
 enum {
     PROP_0,
     PROP_TYPE,
-    PROP_HOST_IP,
+    PROP_ADDRESS,
     PROP_TTL,
     PROP_PORT,
-    PROP_IFA_NAME
+    PROP_IFA_NAME,
+    PROP_IFA_IDX
 };
 
 static void
 gssdp_socket_source_init (GSSDPSocketSource *self)
 {
-        self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
-                                                  GSSDP_TYPE_SOCKET_SOURCE,
-                                                  GSSDPSocketSourcePrivate);
 }
 
 static gboolean
@@ -105,24 +118,29 @@ gssdp_socket_source_set_property (GObject          *object,
                                   GParamSpec       *pspec)
 {
         GSSDPSocketSource *self;
+        GSSDPSocketSourcePrivate *priv;
 
         self = GSSDP_SOCKET_SOURCE (object);
+        priv = gssdp_socket_source_get_instance_private (self);
 
         switch (property_id) {
         case PROP_TYPE:
-                self->priv->type = g_value_get_int (value);
-                break;
-        case PROP_HOST_IP:
-                self->priv->host_ip = g_value_dup_string (value);
+                priv->type = g_value_get_int (value);
                 break;
         case PROP_IFA_NAME:
-                self->priv->device_name = g_value_dup_string (value);
+                priv->device_name = g_value_dup_string (value);
+                break;
+        case PROP_ADDRESS:
+                priv->address = g_value_dup_object (value);
                 break;
         case PROP_TTL:
-                self->priv->ttl = g_value_get_uint (value);
+                priv->ttl = g_value_get_uint (value);
                 break;
         case PROP_PORT:
-                self->priv->port = g_value_get_uint (value);
+                priv->port = g_value_get_uint (value);
+                break;
+        case PROP_IFA_IDX:
+                priv->index = g_value_get_int (value);
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -137,9 +155,10 @@ gssdp_socket_source_set_property (GObject          *object,
  **/
 GSSDPSocketSource *
 gssdp_socket_source_new (GSSDPSocketSourceType type,
-                         const char           *host_ip,
+                         GInetAddress         *address,
                          guint                 ttl,
                          const char           *device_name,
+                         guint                 index,
                          GError              **error)
 {
         return g_initable_new (GSSDP_TYPE_SOCKET_SOURCE,
@@ -147,12 +166,14 @@ gssdp_socket_source_new (GSSDPSocketSourceType type,
                                error,
                                "type",
                                type,
-                               "host-ip",
-                               host_ip,
+                               "address",
+                               address,
                                "ttl",
                                ttl,
                                "device-name",
                                device_name,
+                               "index",
+                               index,
                                NULL);
 }
 
@@ -162,46 +183,41 @@ gssdp_socket_source_do_init (GInitable                   *initable,
                              GError                     **error)
 {
         GSSDPSocketSource *self = NULL;
-        GInetAddress *iface_address = NULL;
+        GSSDPSocketSourcePrivate *priv = NULL;
         GSocketAddress *bind_address = NULL;
         GInetAddress *group = NULL;
         GError *inner_error = NULL;
         GSocketFamily family;
         gboolean success = FALSE;
+        gboolean link_local = FALSE;
 
         self = GSSDP_SOCKET_SOURCE (initable);
-        iface_address = g_inet_address_new_from_string (self->priv->host_ip);
-        if (iface_address == NULL) {
-                g_set_error (error,
-                             GSSDP_ERROR,
-                             GSSDP_ERROR_FAILED,
-                             "Invalid host ip: %s",
-                             self->priv->host_ip);
+        priv = gssdp_socket_source_get_instance_private (self);
 
-                goto error;
-        }
-
-        family = g_inet_address_get_family (iface_address);
+        family = g_inet_address_get_family (priv->address);
 
         if (family == G_SOCKET_FAMILY_IPV4)
                 group = g_inet_address_new_from_string (SSDP_ADDR);
         else {
-                g_set_error_literal (error,
-                                     GSSDP_ERROR,
-                                     GSSDP_ERROR_FAILED,
-                                     "IPv6 address");
-
-                goto error;
+                /* IPv6 */
+                /* According to Annex.A, we need to check the scope of the
+                 * address to use the proper multicast group */
+                if (g_inet_address_get_is_link_local (priv->address)) {
+                            group = g_inet_address_new_from_string (SSDP_V6_LL);
+                            link_local = TRUE;
+                } else {
+                            group = g_inet_address_new_from_string (SSDP_V6_SL);
+                }
         }
 
 
         /* Create socket */
-        self->priv->socket = g_socket_new (G_SOCKET_FAMILY_IPV4,
-                                           G_SOCKET_TYPE_DATAGRAM,
-                                           G_SOCKET_PROTOCOL_UDP,
-                                           &inner_error);
+        priv->socket = g_socket_new (family,
+                                     G_SOCKET_TYPE_DATAGRAM,
+                                     G_SOCKET_PROTOCOL_UDP,
+                                     &inner_error);
 
-        if (!self->priv->socket) {
+        if (!priv->socket) {
                 g_propagate_prefixed_error (error,
                                             inner_error,
                                             "Could not create socket");
@@ -210,9 +226,9 @@ gssdp_socket_source_do_init (GInitable                   *initable,
         }
 
         /* Enable broadcasting */
-        g_socket_set_broadcast (self->priv->socket, TRUE);
+        g_socket_set_broadcast (priv->socket, TRUE);
 
-        if (!gssdp_socket_enable_info (self->priv->socket,
+        if (!gssdp_socket_enable_info (priv->socket,
                                        TRUE,
                                        &inner_error)) {
                 g_propagate_prefixed_error (error,
@@ -223,31 +239,26 @@ gssdp_socket_source_do_init (GInitable                   *initable,
         }
 
         /* TTL */
-        if (!self->priv->ttl)
+        if (priv->ttl == 0) {
                 /* UDA/1.0 says 4, UDA/1.1 says 2 */
-                self->priv->ttl = 4;
+                priv->ttl = 4;
+                if (family == G_SOCKET_FAMILY_IPV6) {
+                        /* UDA 2.0, Annex A says 10 hops */
+                        priv->ttl = 10;
+                }
+        }
 
-        g_socket_set_multicast_ttl (self->priv->socket, self->priv->ttl);
+        g_socket_set_multicast_ttl (priv->socket, priv->ttl);
 
 
         /* Set up additional things according to the type of socket desired */
-        if (self->priv->type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
+        if (priv->type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
                 /* Enable multicast loopback */
-                g_socket_set_multicast_loopback (self->priv->socket, TRUE);
+                g_socket_set_multicast_loopback (priv->socket, TRUE);
 
-
-#ifdef G_OS_WIN32
-                bind_address = g_inet_socket_address_new (iface_address,
-                                                          SSDP_PORT);
-#else
-                bind_address = g_inet_socket_address_new (group,
-                                                          SSDP_PORT);
-#endif
-        } else {
-                guint port = SSDP_PORT;
-
-                if (!gssdp_socket_mcast_interface_set (self->priv->socket,
-                                                       iface_address,
+                if (!gssdp_socket_mcast_interface_set (priv->socket,
+                                                       priv->address,
+                                                       (guint32) priv->index,
                                                        &inner_error)) {
                         g_propagate_prefixed_error (
                                         error,
@@ -257,13 +268,35 @@ gssdp_socket_source_do_init (GInitable                   *initable,
                         goto error;
                 }
 
+#ifdef G_OS_WIN32
+                bind_address = g_inet_socket_address_new (priv->address,
+                                                          SSDP_PORT);
+#else
+                bind_address = g_object_new (G_TYPE_INET_SOCKET_ADDRESS,
+                                             "address", group,
+                                             "port", SSDP_PORT,
+                                             "scope-id", priv->index,
+                                             NULL);
+#endif
+        } else {
+                guint port = SSDP_PORT;
+
                 /* Use user-supplied or random port for the socket source used
                  * by M-SEARCH */
-                if (self->priv->type == GSSDP_SOCKET_SOURCE_TYPE_SEARCH)
-                        port = self->priv->port;
+                if (priv->type == GSSDP_SOCKET_SOURCE_TYPE_SEARCH)
+                        port = priv->port;
+
+                if (link_local) {
+                    bind_address = g_object_new (G_TYPE_INET_SOCKET_ADDRESS,
+                                                 "address", priv->address,
+                                                 "port", port,
+                                                 "scope-id", priv->index,
+                                                 NULL);
+                } else {
+                    bind_address = g_inet_socket_address_new (priv->address,
+                                                              port);
+                }
 
-                bind_address = g_inet_socket_address_new (iface_address,
-                                                          port);
         }
 
         /* Normally g_socket_bind does this, but it is disabled on
@@ -271,7 +304,7 @@ gssdp_socket_source_do_init (GInitable                   *initable,
          * there, also we nees SO_REUSEPORT on OpenBSD. This is a nop
          * everywhere else.
          */
-        if (!gssdp_socket_reuse_address (self->priv->socket,
+        if (!gssdp_socket_reuse_address (priv->socket,
                                          TRUE,
                                          &inner_error)) {
                 g_propagate_prefixed_error (
@@ -283,7 +316,7 @@ gssdp_socket_source_do_init (GInitable                   *initable,
         }
 
         /* Bind to requested port and address */
-        if (!g_socket_bind (self->priv->socket,
+        if (!g_socket_bind (priv->socket,
                             bind_address,
                             TRUE,
                             &inner_error)) {
@@ -294,14 +327,14 @@ gssdp_socket_source_do_init (GInitable                   *initable,
                 goto error;
         }
 
-        if (self->priv->type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
+        if (priv->type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
                 /* The 4th argument 'iface_name' can't be NULL even though Glib API doc says you
                  * can. 'NULL' will fail the test.
                  */
-                if (!g_socket_join_multicast_group (self->priv->socket,
+                if (!g_socket_join_multicast_group (priv->socket,
                                                     group,
                                                     FALSE,
-                                                    self->priv->device_name,  /*   e.g. 'lo' */
+                                                    priv->device_name,  /*   e.g. 'lo' */
                                                     &inner_error)) {
                         char *address = g_inet_address_to_string (group);
                         g_propagate_prefixed_error (error,
@@ -314,14 +347,12 @@ gssdp_socket_source_do_init (GInitable                   *initable,
                 }
         }
 
-        self->priv->source = g_socket_create_source (self->priv->socket,
-                                                     G_IO_IN | G_IO_ERR,
-                                                     NULL);
+        priv->source = g_socket_create_source (priv->socket,
+                                               G_IO_IN | G_IO_ERR,
+                                               NULL);
         success = TRUE;
 
 error:
-        if (iface_address != NULL)
-                g_object_unref (iface_address);
         if (bind_address != NULL)
                 g_object_unref (bind_address);
         if (group != NULL)
@@ -339,9 +370,11 @@ error:
 GSocket *
 gssdp_socket_source_get_socket (GSSDPSocketSource *socket_source)
 {
+        GSSDPSocketSourcePrivate *priv;
         g_return_val_if_fail (socket_source != NULL, NULL);
+        priv = gssdp_socket_source_get_instance_private (socket_source);
 
-        return socket_source->priv->socket;
+        return priv->socket;
 }
 
 void
@@ -349,19 +382,23 @@ gssdp_socket_source_set_callback (GSSDPSocketSource *self,
                                   GSourceFunc        callback,
                                   gpointer           user_data)
 {
+        GSSDPSocketSourcePrivate *priv;
         g_return_if_fail (self != NULL);
         g_return_if_fail (GSSDP_IS_SOCKET_SOURCE (self));
+        priv = gssdp_socket_source_get_instance_private (self);
 
-        g_source_set_callback (self->priv->source, callback, user_data, NULL);
+        g_source_set_callback (priv->source, callback, user_data, NULL);
 }
 
 void
 gssdp_socket_source_attach (GSSDPSocketSource *self)
 {
+        GSSDPSocketSourcePrivate *priv;
         g_return_if_fail (self != NULL);
         g_return_if_fail (GSSDP_IS_SOCKET_SOURCE (self));
+        priv = gssdp_socket_source_get_instance_private (self);
 
-        g_source_attach (self->priv->source,
+        g_source_attach (priv->source,
                          g_main_context_get_thread_default ());
 }
 
@@ -369,19 +406,21 @@ static void
 gssdp_socket_source_dispose (GObject *object)
 {
         GSSDPSocketSource *self;
+        GSSDPSocketSourcePrivate *priv;
 
         self = GSSDP_SOCKET_SOURCE (object);
+        priv = gssdp_socket_source_get_instance_private (self);
 
-        if (self->priv->source != NULL) {
-                g_source_destroy (self->priv->source);
-                g_source_unref (self->priv->source);
-                self->priv->source = NULL;
+        if (priv->source != NULL) {
+                g_source_destroy (priv->source);
+                g_source_unref (priv->source);
+                priv->source = NULL;
         }
 
-        if (self->priv->socket != NULL) {
-                g_socket_close (self->priv->socket, NULL);
-                g_object_unref (self->priv->socket);
-                self->priv->socket = NULL;
+        if (priv->socket != NULL) {
+                g_socket_close (priv->socket, NULL);
+                g_object_unref (priv->socket);
+                priv->socket = NULL;
         }
 
         G_OBJECT_CLASS (gssdp_socket_source_parent_class)->dispose (object);
@@ -391,17 +430,16 @@ static void
 gssdp_socket_source_finalize (GObject *object)
 {
         GSSDPSocketSource *self;
+        GSSDPSocketSourcePrivate *priv;
 
         self = GSSDP_SOCKET_SOURCE (object);
+        priv = gssdp_socket_source_get_instance_private (self);
 
-        if (self->priv->host_ip != NULL) {
-                g_free (self->priv->host_ip);
-                self->priv->host_ip = NULL;
-        }
+        g_clear_object (&priv->address);
 
-        if (self->priv->device_name != NULL) {
-                g_free (self->priv->device_name);
-                self->priv->device_name = NULL;
+        if (priv->device_name != NULL) {
+                g_free (priv->device_name);
+                priv->device_name = NULL;
         }
 
         G_OBJECT_CLASS (gssdp_socket_source_parent_class)->finalize (object);
@@ -419,8 +457,6 @@ gssdp_socket_source_class_init (GSSDPSocketSourceClass *klass)
         object_class->dispose = gssdp_socket_source_dispose;
         object_class->finalize = gssdp_socket_source_finalize;
 
-        g_type_class_add_private (klass, sizeof (GSSDPSocketSourcePrivate));
-
         g_object_class_install_property
                 (object_class,
                  PROP_TYPE,
@@ -437,12 +473,12 @@ gssdp_socket_source_class_init (GSSDPSocketSourceClass *klass)
 
         g_object_class_install_property
                 (object_class,
-                 PROP_HOST_IP,
-                 g_param_spec_string
-                        ("host-ip",
-                         "Host ip",
+                 PROP_ADDRESS,
+                 g_param_spec_object
+                        ("address",
+                         "Host address",
                          "IP address of associated network interface",
-                         NULL,
+                         G_TYPE_INET_ADDRESS,
                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
                          G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                          G_PARAM_STATIC_BLURB));
@@ -483,4 +519,16 @@ gssdp_socket_source_class_init (GSSDPSocketSourceClass *klass)
                          0,
                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
                          G_PARAM_STATIC_STRINGS));
+
+        g_object_class_install_property
+                (object_class,
+                 PROP_IFA_IDX,
+                 g_param_spec_int
+                        ("index",
+                         "Interface index",
+                         "Interface index of the network device",
+                         -1, G_MAXUINT16,
+                         -1,
+                         G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
+                         G_PARAM_STATIC_STRINGS));
 }
