@@ -24,8 +24,8 @@
 #include <string.h>
 #include <stdlib.h>
 
-#define UI_FILE "gssdp-device-sniffer.ui"
-#define MAX_IP_LEN 16
+#define UI_RESOURCE "/org/gupnp/GSSDP/DeviceSniffer.ui"
+#define MENU_RESOURCE "/org/gupnp/GSSDP/WindowMenu.ui"
 
 static char *interface = NULL;
 
@@ -34,26 +34,20 @@ GSSDPResourceBrowser *resource_browser;
 GSSDPClient *client;
 char *ip_filter = NULL;
 gboolean capture_packets = TRUE;
+gboolean prefer_v6 = FALSE;
 
 GOptionEntry entries[] =
 {
         {"interface", 'i', 0, G_OPTION_ARG_STRING, &interface, "Network interface to listen on", NULL },
+        { "prefer-v6", '6', 0, G_OPTION_ARG_NONE, &prefer_v6, "Prefer IPv6 for the client", NULL },
+        { "prefer-v4", '4', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &prefer_v6, "Prefer IPv4 for the client", NULL },
         { NULL }
 };
 
 void
-on_enable_packet_capture_activate (GtkCheckMenuItem *menuitem,
+on_enable_packet_capture_activate (GtkToggleButton  *menuitem,
                                    gpointer          user_data);
-void
-on_details_activate (GtkWidget *scrolled_window, GtkCheckMenuItem *menuitem);
 
-void
-on_clear_packet_capture_activate (GtkMenuItem *menuitem,
-                                  gpointer     user_data);
-
-void
-on_use_filter_radiobutton_toggled (GtkToggleButton *togglebutton,
-                                   gpointer         user_data);
 void
 on_address_filter_dialog_response (GtkDialog *dialog,
                                    gint       response,
@@ -63,10 +57,19 @@ gboolean
 on_delete_event (GtkWidget *widget, GdkEvent  *event, gpointer   user_data);
 
 G_MODULE_EXPORT void
-on_enable_packet_capture_activate (GtkCheckMenuItem      *menuitem,
-                                   G_GNUC_UNUSED gpointer user_data)
+on_enable_packet_capture_activate (GtkToggleButton *menuitem,
+                                   gpointer         user_data)
 {
-        capture_packets = gtk_check_menu_item_get_active (menuitem);
+        const gchar *icon_name = NULL;
+
+        capture_packets = gtk_toggle_button_get_active (menuitem);
+        icon_name = capture_packets
+                ? "media-playback-stop-symbolic"
+                : "media-playback-start-symbolic";
+
+        gtk_image_set_from_icon_name (GTK_IMAGE (user_data),
+                                      icon_name,
+                                      GTK_ICON_SIZE_BUTTON);
 }
 
 static void
@@ -92,14 +95,18 @@ clear_packet_treeview (void)
         }
 }
 
-G_MODULE_EXPORT
-void
-on_details_activate (GtkWidget *scrolled_window, GtkCheckMenuItem *menuitem)
+static void
+on_details_activate (GSimpleAction *action,
+                     GVariant *parameter,
+                     gpointer user_data)
 {
-        gboolean active;
+        GtkWidget *scrolled_window = NULL;
 
-        active = gtk_check_menu_item_get_active (menuitem);
-        g_object_set (G_OBJECT (scrolled_window), "visible", active, NULL);
+        scrolled_window = GTK_WIDGET (gtk_builder_get_object (builder,
+                                                              "packet-details-scrolledwindow"));
+
+        g_object_set (G_OBJECT (scrolled_window), "visible", g_variant_get_boolean (parameter), NULL);
+        g_simple_action_set_state (action, parameter);
 }
 
 static void
@@ -166,8 +173,8 @@ on_packet_selected (GtkTreeSelection      *selection,
 
                 gtk_tree_model_get (model,
                                 &iter, 
-                                4, &packet_headers,
-                                5, &arrival_time, -1);
+                                5, &packet_headers,
+                                6, &arrival_time, -1);
                 display_packet (*arrival_time, packet_headers);
                 g_boxed_free (SOUP_TYPE_MESSAGE_HEADERS, packet_headers);
         }
@@ -176,14 +183,6 @@ on_packet_selected (GtkTreeSelection      *selection,
                 update_packet_details ("", 0);
 }
 
-G_MODULE_EXPORT
-void
-on_clear_packet_capture_activate (G_GNUC_UNUSED GtkMenuItem *menuitem,
-                                  G_GNUC_UNUSED gpointer     user_data)
-{
-        clear_packet_treeview ();
-}
-
 static const char *message_types[] = {"M-SEARCH", "RESPONSE", "NOTIFY"};
 
 static char **
@@ -196,7 +195,7 @@ packet_to_treeview_data (const gchar        *from_ip,
         const char *target;
         struct tm *tm;
 
-        packet_data = g_malloc (sizeof (char *) * 5);
+        packet_data = g_new0 (char *, 6);
 
         /* Set the Time */
         tm = localtime (&arrival_time);
@@ -205,8 +204,11 @@ packet_to_treeview_data (const gchar        *from_ip,
         /* Now the Source Address */
         packet_data[1] = g_strdup (from_ip);
         
+        packet_data[2] = g_strdup (gssdp_client_get_interface (client));
+
+        packet_data[3] = g_strdup ("lo");
         /* Now the Packet Type */
-        packet_data[2] = g_strdup (message_types[type]);
+        packet_data[3] = g_strdup (message_types[type]);
         
         /* Now the Packet Information */
         if (type == _GSSDP_DISCOVERY_RESPONSE)
@@ -214,8 +216,8 @@ packet_to_treeview_data (const gchar        *from_ip,
         else
                 target = soup_message_headers_get_one (headers, "NT");
         
-        packet_data[3] = g_strdup (target);
-        packet_data[4] = NULL;
+        packet_data[4] = g_strdup (target);
+        packet_data[5] = NULL;
 
         return packet_data;
 }
@@ -246,8 +248,9 @@ append_packet (const gchar *from_ip,
                         1, packet_data[1],
                         2, packet_data[2],
                         3, packet_data[3],
-                        4, headers,
-                        5, g_memdup (&arrival_time, sizeof (time_t)),
+                        4, packet_data[4],
+                        5, headers,
+                        6, g_memdup (&arrival_time, sizeof (time_t)),
                         -1);
         g_strfreev (packet_data);
 }
@@ -412,7 +415,7 @@ on_use_filter_radiobutton_toggled (GtkToggleButton       *togglebutton,
         GtkWidget *filter_hbox;
         gboolean use_filter;
 
-        filter_hbox = GTK_WIDGET(gtk_builder_get_object (builder, "address-filter-hbox"));
+        filter_hbox = GTK_WIDGET(gtk_builder_get_object (builder, "address-entry0"));
         g_assert (filter_hbox != NULL);
         
         use_filter = gtk_toggle_button_get_active (togglebutton);
@@ -422,26 +425,20 @@ on_use_filter_radiobutton_toggled (GtkToggleButton       *togglebutton,
 static char *
 get_ip_filter (void)
 {
-        int i;
-        char *ip;
-        guint8 quad[4];
+        GtkEntry *entry;
+        GInetAddress *addr;
 
-        ip = g_malloc (MAX_IP_LEN);
-        for (i=0; i<4; i++) {
-                GtkWidget *entry;
-                char entry_name[16];
-                gint val;
+        entry = GTK_ENTRY (gtk_builder_get_object (builder, "address-entry0"));
+        addr = g_inet_address_new_from_string (gtk_entry_get_text (entry));
 
-                sprintf (entry_name, "address-entry%d", i);
-                entry = GTK_WIDGET(gtk_builder_get_object (builder, entry_name));
-                g_assert (entry != NULL);
+        if (addr == NULL) {
+                g_warning ("Filter not a valid IP address");
 
-                val = atoi (gtk_entry_get_text (GTK_ENTRY (entry)));
-                quad[i] = CLAMP (val, 0, 255);
+                return NULL;
         }
-        sprintf (ip, "%u.%u.%u.%u", quad[0], quad[1], quad[2], quad[3]);
-        
-        return ip;
+        g_object_unref (addr);
+
+        return g_strdup (gtk_entry_get_text (entry));
 }
 
 G_MODULE_EXPORT
@@ -474,7 +471,8 @@ create_packet_treemodel (void)
 {
         GtkListStore *store;
 
-        store = gtk_list_store_new (6,
+        store = gtk_list_store_new (7,
+                        G_TYPE_STRING,
                         G_TYPE_STRING,
                         G_TYPE_STRING,
                         G_TYPE_STRING,
@@ -524,13 +522,24 @@ setup_treeview (GtkWidget *treeview,
                         model);
 }
 
+static gboolean
+on_treeview_popup_menu (GtkWidget *tv, GdkEventButton *event, gpointer user_data)
+{
+        if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
+                gtk_menu_popup_at_pointer (GTK_MENU (user_data), (GdkEvent *)event);
+        }
+
+        return FALSE;
+}
+
 static void
 setup_treeviews (void)
 {
         GtkWidget *treeviews[2];
         GtkTreeModel *treemodels[2];
-        const char *headers[2][6] = { {"Time",
+        const char *headers[2][7] = { {"Time",
                 "Source Address",
+                "Interface",
                 "Packet Type",
                 "Packet Information",
                 NULL }, {"Unique Identifier",
@@ -539,6 +548,7 @@ setup_treeviews (void)
                 "Location",
                 NULL } }; 
         GtkTreeSelection *selection;
+        GtkWidget *menu = NULL;
         int i;
 
         treeviews[0] = GTK_WIDGET(gtk_builder_get_object (builder,
@@ -562,6 +572,12 @@ setup_treeviews (void)
                         "changed",
                         G_CALLBACK (on_packet_selected),
                         (gpointer *) treeviews[0]);
+        menu = gtk_menu_new_from_model (G_MENU_MODEL (gtk_builder_get_object (builder,
+                                                                              "sniffer-context-menu")));
+        gtk_menu_attach_to_widget (GTK_MENU (menu), treeviews[0], NULL);
+        g_signal_connect (G_OBJECT (treeviews[0]),
+                          "button-press-event",
+                          G_CALLBACK (on_treeview_popup_menu), menu);
 }
 
 G_MODULE_EXPORT
@@ -575,6 +591,78 @@ on_delete_event (G_GNUC_UNUSED GtkWidget *widget,
         return TRUE;
 }
 
+static void
+on_show_address_filter (GSimpleAction *action,
+                        GVariant *parameter,
+                        gpointer user_data)
+{
+        GtkWidget *dialog = NULL;
+
+        dialog = GTK_WIDGET (gtk_builder_get_object (builder,
+                                                     "address-filter-dialog"));
+        gtk_widget_show (dialog);
+}
+
+static void
+on_set_address_filter (GSimpleAction *action,
+                       GVariant      *parameter,
+                       gpointer       user_data)
+{
+        GtkWidget *treeview = NULL;
+        GtkTreeSelection *selection = NULL;
+        GtkTreeModel *model = NULL;
+        GtkTreeIter iter;
+        GtkWidget *use_filter_radio;
+
+        treeview = GTK_WIDGET (gtk_builder_get_object (builder, "packet-treeview"));
+        use_filter_radio = GTK_WIDGET(gtk_builder_get_object (builder, "use-filter-radiobutton"));
+
+        selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
+        gtk_tree_selection_get_selected (selection, &model, &iter);
+        g_free (ip_filter);
+        gtk_tree_model_get (model, &iter,
+                        1, &ip_filter, -1);
+        gtk_entry_set_text (GTK_ENTRY (gtk_builder_get_object (builder, "address-entry0")),
+                            ip_filter);
+
+        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (use_filter_radio), TRUE);
+}
+
+static void
+on_about (GSimpleAction *action, GVariant *parameter, gpointer user_data)
+{
+        GtkWidget *dialog = NULL;
+
+        dialog = GTK_WIDGET (gtk_builder_get_object (builder,
+                                                     "about-dialog"));
+        gtk_widget_show (dialog);
+}
+
+
+G_MODULE_EXPORT
+void
+on_clear_packet_capture_activate (G_GNUC_UNUSED GtkMenuItem *menuitem,
+                                  G_GNUC_UNUSED gpointer     user_data)
+{
+        clear_packet_treeview ();
+}
+
+static void
+on_trigger_rescan (GSimpleAction *action,
+                   GVariant *parameter,
+                   gpointer user_data)
+{
+        gssdp_resource_browser_rescan (resource_browser);
+}
+
+static GActionEntry actions[] = {
+        { "trigger-rescan", on_trigger_rescan },
+        { "set-address-filter", on_set_address_filter },
+        { "show-packet-details", NULL, NULL, "true", on_details_activate },
+        { "show-address-filter", on_show_address_filter },
+        { "about", on_about }
+};
+
 static gboolean
 init_ui (gint *argc, gchar **argv[])
 {
@@ -583,6 +671,7 @@ init_ui (gint *argc, gchar **argv[])
         GError *error = NULL;
         GOptionContext *context;
         double w, h;
+        GSimpleActionGroup *group = NULL;
 
         context = g_option_context_new ("- graphical SSDP debug tool");
         g_option_context_add_main_entries (context, entries, NULL);
@@ -595,26 +684,22 @@ init_ui (gint *argc, gchar **argv[])
         }
 
         builder = gtk_builder_new();
+        if (gtk_builder_add_from_resource(builder, UI_RESOURCE, NULL) == 0)
+                return FALSE;
 
-        /* Try to fetch the ui file from the CWD first */
-        if (gtk_builder_add_from_file (builder, UI_FILE, NULL) == 0) {
-            /* Apparently not. let's check next to the executable */
-            char *path = g_strconcat (g_path_get_dirname (*argv[0]), G_DIR_SEPARATOR_S, UI_FILE, NULL);
-            if (gtk_builder_add_from_file (builder, path, NULL) == 0) {
-                g_clear_pointer (&path, g_free);
-                /* Also not... Check the install path */
-                if (gtk_builder_add_from_file (builder, UI_DIR G_DIR_SEPARATOR_S UI_FILE, NULL) == 0) {
-                    g_critical ("Unable to load the GUI file %s", UI_FILE);
+        if (gtk_builder_add_from_resource (builder, MENU_RESOURCE, NULL) == 0)
+                return FALSE;
 
-                    return FALSE;
-                }
-            }
-            g_clear_pointer (&path, g_free);
-        }
+        gtk_menu_button_set_menu_model (GTK_MENU_BUTTON (gtk_builder_get_object (builder, "window-menu")),
+                        G_MENU_MODEL (gtk_builder_get_object (builder, "sniffer-window-menu")));
 
         main_window = GTK_WIDGET(gtk_builder_get_object (builder, "main-window"));
         g_assert (main_window != NULL);
 
+        group = g_simple_action_group_new ();
+        gtk_widget_insert_action_group (GTK_WIDGET (main_window), "win", G_ACTION_GROUP (group));
+        g_action_map_add_action_entries (G_ACTION_MAP (group), actions, G_N_ELEMENTS (actions), NULL);
+
 
 #if GTK_CHECK_VERSION(3,22,0)
         gtk_widget_realize (main_window);
@@ -663,6 +748,7 @@ init_upnp (void)
         client = g_initable_new (GSSDP_TYPE_CLIENT,
                                  NULL,
                                  &error,
+                                 "host-ip", prefer_v6 ? "::" : NULL,
                                  "interface", interface,
                                  NULL);
         if (error) {
